/*
 *  TOPPERS/SSP Kernel
 *      Smallest Set Profile Kernel
 *
 *  Copyright (C) 2008 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2010 by Meika Sugimoto
 *  Copyright (C) 2018 by Naoki Saito
 *             Nagoya Municipal Industrial Research Institute, JAPAN
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: prc_config.c 1304 2008-08-27 07:28:36Z ertl-honda $
 */

/*
 *		プロセッサ依存モジュール（ARM64用）
 */

#include "kernel_impl.h"
#include <t_syslog.h>

/* 割込みネスト数の管理 */
uint8_t intnest;

/*
 *  プロセッサ依存の初期化
 */
void
prc_initialize(void)
{
    /* 割込みネスト数の初期化 */
    intnest = 0u;

    /*
     *  GICの設定
     */
    gicd_initialize();
    gicc_initialize();
}

/*
 *  プロセッサ依存の終了処理
 */
void
prc_terminate(void)
{
    while(1)
    ;
}

/*
 *  割込み要求ライン属性の設定
 */
void
x_config_int(INTNO intno, ATR intatr, PRI intpri)
{
    assert(VALID_INTNO(intno));
    assert(TMIN_INTPRI <= intpri && intpri <= TMAX_INTPRI);

    /*
     *  割込みを禁止
     *
     *  割込みを受け付けたまま，レベルトリガ／エッジトリガの設定や，割
     *  込み優先度の設定を行うのは危険なため，割込み属性にかかわらず，
     *  一旦マスクする．
     */
    x_disable_int(intno);

    /*
     *  割込みをコンフィギュレーション
     */

    if ((intatr & TA_EDGE) != 0U) {
        gicd_config(intno, GICD_ICFGRn_EDGE);
        x_clear_int(intno);
    }
    else {
        gicd_config(intno, GICD_ICFGRn_LEVEL);
    }

    /*
     *  割込み優先度とターゲットプロセッサを設定
     */
    gicd_set_priority(intno, INT_IPM(intpri));
    gicd_set_target(intno, 1U << ARM_PRC_INDEX);

    /*
     * 割込みを許可
     */
    if ((intatr & TA_ENAINT) != 0U) {
        x_enable_int(intno);
    }
}

#ifndef OMIT_DEFAULT_EXC_HANDLER
/*
 *  登録されていない例外が発生すると呼び出される
 */
void
default_exc_handler(void *p_excinf)
{
    uint64_t gicc_pmr = *(((uint64_t*)p_excinf) + 0);
    uint64_t excno    = *(((uint64_t*)p_excinf) + 1);
    // スタックポインタの調整量
    uint64_t adj      = *(((uint64_t*)p_excinf) + 3) >> 3;
    uint64_t esr, pc, spsr, sp;

    esr      = *(((uint64_t*)p_excinf) + 4 + adj);
    pc       = *(((uint64_t*)p_excinf) + 5 + adj);
    spsr     = *(((uint64_t*)p_excinf) + 6 + adj);
    sp       = *(((uint64_t*)p_excinf) + 38 + adj);

    syslog(LOG_EMERG, "\nUnregistered CPU Exception occured.");
    syslog(LOG_EMERG, "EXCNO = %08x ESR = %08x GICC_PMR = %08x ", excno, esr, gicc_pmr);
    syslog(LOG_EMERG, "PC=%08x%08x SP = %08x%08x SPSR = %08x",
        (uint32_t)((pc & UINT64_C(0xffffffff00000000))>>32), (uint32_t)(pc & UINT64_C(0xffffffff)),
        (uint32_t)((sp & UINT64_C(0xffffffff00000000))>>32), (uint32_t)(sp & UINT64_C(0xffffffff)),
        (uint32_t)spsr);

    target_exit();
}
#endif /* OMIT_DEFAULT_EXC_HANDLER */

#ifndef OMIT_DEFAULT_INT_HANDLER
/*
 *  未登録の割込みが発生した場合に呼び出される
 */
void
default_int_handler(void)
{
    syslog(LOG_EMERG, "\nUnregistered Interrupt occurs.");

    target_exit();
}
#endif /* OMIT_DEFAULT_INT_HANDLER */

/*
 *  割込みハンドラの呼び出し
 */
void
call_int_handler(uint32_t intno)
{
    intnest++;

    i_unlock_cpu();

    // ハンドラの呼び出し
    inh_table[intno]();

    i_lock_cpu();

    intnest--;
}

/*
 *  CPU例外ハンドラの呼び出し
 */
void
call_exc_handler(void *p_excinf, uint32_t excno)
{
    intnest++;

    // CPU例外発生時はCPUロックフラグを操作しない

    // ハンドラの呼び出し
    exc_table[excno](p_excinf);

    i_lock_cpu();

    intnest--;
}

