/*
 *  TOPPERS/SSP Kernel
 *      Smallest Set Profile Kernel
 *
 *  Copyright (C) 2013-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2018 by Naoki Saito
 *             Nagoya Municipal Industrial Research Institute, JAPAN
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  $Id$
 */

/*
 *  タイマドライバ（Generic Timer）
 */
#include "kernel_impl.h"
#include "time_event.h"
#include "target_timer.h"
#include <sil.h>

/*
 *  仮想タイマの制御レジスタからの読み出し
 */
Inline uint64_t
clock_read_ctrl(void) {
    uint64_t cntv_ctrl;

    Asm ("isb; mrs %0, cntv_ctl_el0; isb;"
        : "=r"(cntv_ctrl) :: "memory");

    return cntv_ctrl;
}

/*
 *  仮想タイマの制御レジスタへの書き込み
 */
Inline void
clock_write_ctrl(uint64_t cntv_ctrl) {

    Asm ("isb; msr cntv_ctl_el0, %0; isb;"
        :: "r"(cntv_ctrl) : "memory");
    return ;
}

/*
 *  仮想タイマ(フリーカウンタ)のカウント値の読み出し
 */
Inline uint64_t
clock_read_counter(void){
    uint64_t cval;

    Asm ("mrs %0, cntvct_el0" : "=r" (cval));

    return cval;
}

/*
 *  仮想タイマの値の読み出し
 */
Inline uint64_t
clock_read_tval(void){
    uint64_t cntv_tval;

    Asm ("isb; mrs %0, cntv_tval_el0; isb;" : "=r"(cntv_tval)
        :: "memory");
    return cntv_tval;
}

/*
 *  仮想タイマの値の書き込み
 */
Inline void
clock_write_tval(uint64_t cntv_tval){

    Asm ("isb; msr cntv_tval_el0, %0; isb;" :: "r"(cntv_tval)
        : "memory");
}

/*
 *  システムカウンタのクロック周波数の取得
 */
Inline uint64_t
get_timer_freq(void) {
    uint64_t freq_hz;

    Asm ("mrs %0, cntfrq_el0; isb; " : "=r"(freq_hz) :: "memory");
    return freq_hz;
}

/*
 *  タイマの開始
 */
static void
target_timer_start(void)
{
    uint64_t cntv_ctl;

    cntv_ctl = clock_read_ctrl();
    cntv_ctl |=  ( CNTV_CTL_ENABLE | CNTV_CTL_START );
    cntv_ctl &= ~CNTV_CTL_MASK; //mask
    clock_write_ctrl(cntv_ctl);
}

/*
 *  タイマの停止
 */
static void
target_timer_stop(void)
{
    uint64_t cntv_ctl;

    cntv_ctl = clock_read_ctrl();
    cntv_ctl &= ~( CNTV_CTL_ENABLE | CNTV_CTL_START );
    cntv_ctl |=  CNTV_CTL_MASK;
    clock_write_ctrl(cntv_ctl);
}

/*
 *  タイマ割込み要求のクリア
 */
static void
target_timer_int_clear(void){

    gicd_clear_pending(INTNO_TIMER);
}

/*
 *  次のタイマ割込み発生までのオフセット
 */
static uint64_t next_tval;

/*
 *  タイマの起動処理
 */
void
target_timer_initialize(intptr_t exinf)
{
    target_timer_stop();
    target_timer_int_clear();

    // 次回のタイマ割込み発生までのオフセット算出
    next_tval = get_timer_freq() * 1000 / 1000000UL;  /* 1ms */
    clock_write_tval(next_tval);

    target_timer_start();
}

/*
 *  タイマの停止処理
 */
void
target_timer_terminate(intptr_t exinf)
{
    target_timer_stop();
}

/*
 *  タイマ割込みハンドラ
 */
void
target_timer_handler(void)
{
    target_timer_stop();
    target_timer_int_clear();
    clock_write_tval(next_tval);    // 次回の割込み
    target_timer_start();

    i_begin_int(INTNO_TIMER);
    signal_time();
    i_end_int(INTNO_TIMER);
}
